@@ -4,6 +4,8 @@
  *
  * mac: Mac partition structures
  *
+ * This file is part of mmtools
+ *
  * mmtools is free software; you can redistribute it and/or modify it
  * under the terms of the GNU General Public License as published by the
  * Free Software Foundation; either version 2 of the License, or (at your
@@ -50,7 +52,7 @@ mac_load_table(MM_INFO *mm, daddr_t sect)
 	off_t ret;
 	char *table_str;
 	off_t addr = (off_t)sect * 512;
-	u_int32_t idx = 0, max_part = 0;
+	u_int32_t idx, max_part;
 
 	if (verbose)
 		fprintf(logfp,
@@ -66,7 +68,9 @@ mac_load_table(MM_INFO *mm, daddr_t sect)
 	 * The idx variable shows which round it is
 	 * Each structure is 512-bytes each
 	 */
-	do {
+
+	max_part = 1;  /* set it to 1 and it will be set in the first loop */
+	for (idx = 0; idx < max_part; idx++) {
 		u_int32_t part_start;
 		u_int32_t part_size;
 		char *str;
@@ -102,20 +106,18 @@ mac_load_table(MM_INFO *mm, daddr_t sect)
 			  "mac_load: %d  Starting Sector: %lu  Size: %lu Type: %s\n",
 			  idx, (ULONG)part_start, (ULONG)part_size, part.type);
 
-		if (part_size == 0)
-			continue;
+		if (part_size != 0) {
 
-		str = mymalloc(sizeof(part.name));
-		strncpy(str, part.type, sizeof(part.name));
+			str = mymalloc(sizeof(part.name));
+			strncpy(str, part.type, sizeof(part.name));
 
-		mm_part_add(mm, part_start, part_size, 
+			mm_part_add(mm, part_start, part_size, 
 				str, -1, idx);
-
-		idx++;
-	} while (idx < max_part);
+		}
+	}
 
 	/* Add an entry for the table length */
-    table_str = mymalloc(16);
+	table_str = mymalloc(16);
 	snprintf(table_str, 16, "Table");
 	mm_part_add(mm, sect, max_part, table_str, -1, -1);
 	return;
