@@ -45,6 +45,11 @@ extern FILE *logfp;
 #include <sys/sysmacros.h>
 #define STRTOUL		strtoul
 
+
+/* Sol 5.7 has inttypes, but sys/inttypes is needed for PRI.. macros */
+#include <inttypes.h>
+#include <sys/inttypes.h>
+
 #define u_int8_t	uint8_t
 #define u_int16_t	uint16_t
 #define u_int32_t	uint32_t
@@ -76,6 +81,12 @@ extern int optind;
 #include <sys/uio.h>
 #define SUPPORTED
 #define STRTOUL		strtoul
+
+/* FreeBSD 5 has inttypes and support for the printf macros */
+#if defined(FREEBSD4) || defined(FREEBSD5)
+#include <inttypes.h>
+#endif
+
 #endif /* FREEBSD */
 
  /*
@@ -84,6 +95,7 @@ extern int optind;
 #if defined(BSDI2) || defined(BSDI3) || defined(BSDI4)
 #define SUPPORTED
 #define STRTOUL		strtoul
+#include <inttypes.h>
 #endif /* BSDI */
 
  /*
@@ -92,11 +104,13 @@ extern int optind;
 #if defined(OPENBSD2) || defined(OPENBSD3)
 #define SUPPORTED
 #define STRTOUL		strtoul
+#include <inttypes.h>
 #endif
 
 #if defined(DARWIN)
 #define SUPPORTED
 #define STRTOUL		strtoul
+#include <inttypes.h>
 #endif /* DARWIN */
 
 
@@ -106,16 +120,19 @@ extern int optind;
 #if defined(LINUX2) 
 #define SUPPORTED
 #include <linux/types.h>
+#include <inttypes.h>
 #define STRTOUL		strtoul
 #endif /* LINUX */
 
 #if defined(CYGWIN)
+#include <inttypes.h>
 #define SUPPORTED
 #define STRTOUL		strtoul
 #endif /* CYGWIN */
 
 #if defined(__INTERNIX)
 #define SUPPORTED
+#include <inttypes.h>
 #define STRTOUL strtoul
 #define daddr_t long
 #endif
@@ -131,6 +148,57 @@ extern int optind;
 #endif
 
 
+/* printf macros - if the OS doesnot have inttypes.h yet */
+
+#ifndef PRIx64
+#define PRIx64 "llx"
+#endif
+
+#ifndef PRIX64
+#define PRIX64 "llX"
+#endif
+
+#ifndef PRIu64
+#define PRIu64 "llu"
+#endif
+
+#ifndef PRId64
+#define PRId64 "lld"
+#endif
+
+#ifndef PRIo64
+#define PRIo64 "llo"
+#endif
+
+#ifndef PRIx32
+#define PRIx32 "x"
+#endif
+
+#ifndef PRIX32
+#define PRIX32 "X"
+#endif
+
+#ifndef PRIu32
+#define PRIu32 "u"
+#endif
+
+
+
+#ifndef PRIx16
+#define PRIx16 "hx"
+#endif
+
+#ifndef PRIX16
+#define PRIX16 "hX"
+#endif
+
+#ifndef PRIu16
+#define PRIu16 "hu"
+#endif
+
+
+
+
 typedef unsigned long ULONG;
 
 /* Flags for the return value of walk actions */
